"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const fs = __importStar(require("fs"));
const util_1 = require("util");
const os_lock_1 = require("os-lock");
const utils_1 = require("./utils");
const fsOpen = util_1.promisify(fs.open);
const fsClose = util_1.promisify(fs.close);
const fsWrite = util_1.promisify(fs.write);
class File extends events_1.EventEmitter {
    constructor(filename) {
        super();
        this.filename = filename;
        this.fd = null;
        this.lockedPositions = new Map();
        this.logger = utils_1.logger('file');
        this.reads = 0;
        this.writes = 0;
    }
    get isOpen() {
        return this.fd != null;
    }
    read(position, reverse = false, buffer) {
        ++this.reads;
        if (!this.fd)
            throw new Error('Need to call open() before read()');
        return utils_1.read(this.fd, position, reverse, buffer);
    }
    async write(position, buffer) {
        ++this.writes;
        await fsWrite(this.fd, buffer, 0, buffer.length, position);
    }
    async clear(position, length, char = ' ') {
        const buffer = Buffer.alloc(length, char);
        await fsWrite(this.fd, buffer, 0, length, position);
    }
    async truncate(position) {
        await util_1.promisify(fs.ftruncate)(this.fd, position);
    }
    async exists() {
        return await util_1.promisify(fs.exists)(this.filename);
    }
    async append(text) {
        await util_1.promisify(fs.appendFile)(this.filename, text);
    }
    async delete() {
        await util_1.promisify(fs.unlink)(this.filename);
    }
    async stat() {
        return await util_1.promisify(fs.stat)(this.filename);
    }
    async open(mode = 'r+') {
        this.logger.log('opening', this.filename);
        if (this.isOpen)
            throw new Error('File already open');
        this.fd = await fsOpen(this.filename, mode);
    }
    openSync(mode = 'r+') {
        if (this.isOpen)
            throw new Error('File already open');
        this.fd = fs.openSync(this.filename, mode);
    }
    async close() {
        this.logger.log('closing', this.filename);
        this.logger.log('reads', this.reads);
        this.logger.log('writes', this.writes);
        if (!this.fd)
            throw new Error('No open file to close');
        await fsClose(this.fd);
        this.fd = null;
        this.reads = 0;
        this.writes = 0;
    }
    closeSync() {
        if (!this.fd)
            throw new Error('No open file to close');
        fs.closeSync(this.fd);
    }
    async lock(pos = 0, options = { exclusive: false }) {
        const lockedPosition = this.lockedPositions.get(pos) ||
            { count: 0, exclusive: false };
        const canGetLock = options.exclusive ?
            !lockedPosition.count : !lockedPosition.exclusive;
        if (canGetLock) {
            ++lockedPosition.count;
            lockedPosition.exclusive = options.exclusive;
            this.lockedPositions.set(pos, lockedPosition);
            await os_lock_1.lock(this.fd, pos, 1, options);
            return;
        }
        else {
            const timeout = setInterval(() => { }, ~0 >>> 1);
            await new Promise(resolve => {
                this.once('unlock', async () => {
                    clearInterval(timeout);
                    await this.lock(pos, options);
                    resolve();
                });
            });
        }
    }
    async unlock(pos = 0) {
        await os_lock_1.unlock(this.fd, pos, 1);
        const lockedPosition = this.lockedPositions.get(pos);
        lockedPosition.exclusive = false;
        --lockedPosition.count;
        if (!lockedPosition.count) {
            this.lockedPositions.delete(pos);
            this.emit('unlock', pos);
        }
    }
}
exports.default = File;
//# sourceMappingURL=file.js.map